// 
//  This is an IDA plugin, which loads a COFF obj/lib file signature...
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the author be held liable for any damages
//  arising from the use of this software.
//
//--------------------------------------------------------------------------
#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <name.hpp>
#include <segment.hpp>
#include <funcs.hpp>
#include <string>
#include <vector>
#include <stdexcept>
#include "coff.h"
//--------------------------------------------------------------------------
using namespace std;
//--------------------------------------------------------------------------
#define _S(x)     (string(x, sizeof(x)).c_str())
//--------------------------------------------------------------------------
#pragma pack(push, 4)

typedef struct  
{
	string name;
	unsigned value;
	short section;
	unsigned short type;
	unsigned char sClass, numAux;
} ge_symbol_t;

typedef struct  
{
	unsigned offset;
	unsigned symTabIndex;
	unsigned short type;
} ge_reloc_t;

typedef struct
{
	coff_section_header section;
	unsigned sectionNum;
} ge_section_t;

#pragma pack(pop)
//--------------------------------------------------------------------------
int idaapi init(void)
{
	if ( inf.filetype != f_PE ) return PLUGIN_SKIP; // only for PE files  
	return PLUGIN_OK;
}

//--------------------------------------------------------------------------
void idaapi term(void)
{
}
//--------------------------------------------------------------------------
static ea_t eaStart = BADADDR;
static ea_t eaEnd   = BADADDR;
//--------------------------------------------------------------------------

void ReadSymbolTable(const coff_file_header & hdr, const coff_symbol * symTab, char * stringTable, vector<ge_symbol_t> & ge_symbols)
{
	char *name = 0;
	char buff[9]={0};
	//buff[8] = '\0';

	//for (unsigned symIdx = 0; symIdx < hdr.num_symbols; symIdx++)

	unsigned symIdx = 0;
	while (symIdx < hdr.num_symbols)
	{
		if( symTab->name.non_name.zeros == 0 ) 
		{
			name = stringTable + symTab->name.non_name.offset - 4;
		} 
		else
		{
			memcpy( buff, symTab->name.name_string, sizeof(symTab->name.name_string) );
			name = buff;
		}

		ge_symbols[symIdx].name = name;
		ge_symbols[symIdx].section = symTab->sec_num;
		ge_symbols[symIdx].type = symTab->type;
		ge_symbols[symIdx].value = symTab->value;
		ge_symbols[symIdx].sClass = symTab->storage_class;
		ge_symbols[symIdx].numAux = symTab->num_aux;

		unsigned num_aux = symTab->num_aux;
		symTab++;
		if (num_aux > 0)
		{
			symIdx += num_aux;
			symTab += num_aux;
		}

		symIdx++;
	}

}

void MakeNamesForSection(int ourSection, ea_t baseAddr, vector<ge_reloc_t> & ge_relocs, vector<ge_symbol_t> & ge_symbols)
{
		for (unsigned k = 0; k < ge_symbols.size(); k++)
		{
			if (ge_symbols[k].section == ourSection && ge_symbols[k].type == 0x0020)
			{
				do_name_anyway(baseAddr + ge_symbols[k].value, ge_symbols[k].name.c_str());

				func_t * f=get_func(baseAddr + ge_symbols[k].value);
				if ( f )
				{
					f->flags|=FUNC_LIB;
					update_func(f);
				}
			}
		}

		for (unsigned k = 0; k < ge_relocs.size(); k++)
		{
			ulong u = get_long(baseAddr + ge_relocs[k].offset);
			ea_t ea = u;
			if (ge_relocs[k].type == IMAGE_REL_I386_REL32)
			{
				ea += (baseAddr + ge_relocs[k].offset + 4);
			}

			const char * relocName = ge_symbols[ ge_relocs[k].symTabIndex ].name.c_str();
			do_name_anyway(ea, relocName);
		}

		refresh_idaview_anyway();

}

ea_t FindBase(FILE * fp, int bookmark, const coff_section_header * textSectionHdr, vector<ge_reloc_t> ge_relocs)
{
	unsigned int codeBlockLen = textSectionHdr->size;

	if (!codeBlockLen)
		return BADADDR;
	
	char * codeBlock = (char*)qalloc(codeBlockLen);
	memset(codeBlock, 0, codeBlockLen);

	qfseek(fp, bookmark + (int)textSectionHdr->rawdata_ptr, SEEK_SET);
	qfread(fp, codeBlock, codeBlockLen);

	vector<unsigned char> mask(codeBlockLen, 1);
	for (unsigned i = 0; i < ge_relocs.size(); i++)
	{
		if (ge_relocs[i].type == IMAGE_REL_I386_DIR32 || ge_relocs[i].type == IMAGE_REL_I386_REL32)
		{
			mask[ ge_relocs[i].offset ] = 0;
			mask[ ge_relocs[i].offset + 1 ] = 0;
			mask[ ge_relocs[i].offset + 2 ] = 0;
			mask[ ge_relocs[i].offset + 3 ] = 0;
		}
	}

	ea_t result= bin_search(eaStart, eaEnd, (unsigned char*)codeBlock, &mask[0], codeBlockLen, BIN_SEARCH_FORWARD, BIN_SEARCH_CASE);

	qfree(codeBlock);

	return result;
}

void ReadRelocs(FILE * fp, int bookmark, const coff_section_header * textSectionHdr, vector<ge_reloc_t> & ge_relocs)
{
	coff_reloc * rel = (coff_reloc *)qalloc(textSectionHdr->num_relocs * sizeof(coff_reloc));
	qfseek(fp, bookmark + (int)textSectionHdr->reloc_ptr, SEEK_SET);
	qfread(fp, rel, sizeof(coff_reloc) * textSectionHdr->num_relocs);

	

	for (unsigned i = 0; i < textSectionHdr->num_relocs; i++)
	{
		ge_relocs[i].offset = rel[i].offset;
		ge_relocs[i].symTabIndex = rel[i].sym_tab_index;
		ge_relocs[i].type = rel[i].type;
	}

	qfree(rel);
}

void ReadSectionHeaders(FILE * fp, int bookmark, const coff_file_header & hdr, vector<ge_section_t> & ge_textSections)
{
	coff_section_header * secHdr = (coff_section_header *)qalloc(hdr.num_sections * sizeof(coff_section_header));
	qfseek(fp, bookmark + (int)sizeof(coff_file_header), SEEK_SET);
	qfread(fp, secHdr, sizeof(coff_section_header) * hdr.num_sections);
	
	for (unsigned i = 0; i < hdr.num_sections; i++)
	{
		if (_memicmp(secHdr->name, ".text", 5) == 0 || ((secHdr[i].flags & IMAGE_SCN_CNT_CODE) == IMAGE_SCN_CNT_CODE))
		{
			ge_section_t sect;
			sect.section = secHdr[i];
			sect.sectionNum = i+1;
			ge_textSections.push_back(sect);
		}
	}

	qfree(secHdr);
}

void ParseSections(FILE * fp, int bookmark, vector<ge_section_t> & ge_textSections, vector<ge_symbol_t> & ge_symbols)
{
	for (unsigned g = 0; g < ge_textSections.size(); g++)
	{

		coff_section_header * textSectionHdr = &ge_textSections[g].section;

		if (!textSectionHdr->size)
			continue;

		vector<ge_reloc_t> ge_relocs(textSectionHdr->num_relocs);
		ReadRelocs(fp, bookmark, textSectionHdr, ge_relocs);

		ea_t baseAddr=FindBase(fp, bookmark, textSectionHdr, ge_relocs);
		if (baseAddr != BADADDR) 
		{
			MakeNamesForSection((int)ge_textSections[g].sectionNum, baseAddr, ge_relocs, ge_symbols);

			msg("COFF text signature found at address: %08a\n", baseAddr);
			msg("COFF Plugin -> SymbolCount: %d, RelocCount: %d\n", ge_symbols.size(), ge_relocs.size());
		}

		if (wasBreak()) 
			break;
	}
}

void ReadSymbols(FILE * fp, int bookmark, const coff_file_header & hdr, vector<ge_symbol_t> & ge_symbols)
{
	unsigned strTablePos = hdr.sym_table + (hdr.num_symbols * sizeof(coff_symbol));
	qfseek(fp, bookmark + (int)strTablePos, SEEK_SET);

	unsigned strTableSize = 0;
	qfread(fp, &strTableSize, sizeof(unsigned));

	if (!strTableSize)
	{
		msg("COFF Plugin -> string table not found in this COFF file!\n");
		return;
	}

	strTableSize -= sizeof(strTablePos);

	char * stringTable = (char *)qalloc(strTableSize); 
	qfread(fp, stringTable, strTableSize);

	qfseek(fp, bookmark + (int)hdr.sym_table, SEEK_SET);

	coff_symbol * symbolTable = (coff_symbol *)qalloc(sizeof(coff_symbol) * hdr.num_symbols);
	qfread(fp, symbolTable, sizeof(coff_symbol) * hdr.num_symbols);

	ReadSymbolTable(hdr, symbolTable, stringTable, ge_symbols);
	qfree(symbolTable);

	qfree(stringTable);
}

void applyCoffObject(FILE * fp)
{
	int bookmark = qftell(fp);
	

	coff_file_header hdr;
	qfread(fp, &hdr, sizeof(coff_file_header));
	if (hdr.cpu_type != IMAGE_FILE_MACHINE_I386)
	{
		msg("COFF Plugin -> Expected 32bit i386 format (format: %02a)!\n", hdr.cpu_type);
		return;
	}

	vector<ge_symbol_t> ge_symbols(hdr.num_symbols);
	ReadSymbols(fp,bookmark,hdr,ge_symbols);

	vector<ge_section_t> ge_textSections;
	ReadSectionHeaders(fp,bookmark,hdr,ge_textSections);
	if (!ge_textSections.size())
	{
		msg("COFF Plugin -> code section not found in this COFF file!\n");
	}
	else
		ParseSections(fp, bookmark, ge_textSections, ge_symbols);

}
//--------------------------------------------------------------------------
inline void align(FILE * f)
{
	if (qftell(f) % 2 == 1) 
		qfseek(f, +1, SEEK_CUR);
}
//--------------------------------------------------------------------------
bool LibSignatureOk(FILE *fp)
{
	char buf[10]={0};
	qfread(fp, buf, 8);

	if (strncmp(buf, arch_signature, 8) != 0)
	{
		error("This file is not a lib file!");
		return false;
	}

	return true;
}

void ParseLibFile(FILE * fp)
{
		
	arch_header head;
	int bookmark_head;
	static int header_size = sizeof(arch_header);
	qfread(fp, &head, header_size);
	while (qfseek(fp, atol(_S(head.size)), SEEK_CUR) == 0)
	{
		align(fp);
		if (qfread(fp, &head, header_size) == header_size) {
			bookmark_head = qftell(fp);
			applyCoffObject(fp);
		}
		else
		{
			msg("COFF Plugin -> end of file reached!\n");
			break;
		}
		if (wasBreak()) 
			break;
		qfseek(fp, bookmark_head, SEEK_SET);
	}
	//else
	//	error("This lib file has no object files within!");

	
}

void idaapi run(int )
{
	const char * filename = askfile_c(0,"*.obj;*.lib", "Please select COFF object file to find signature in...");
	if (filename == NULL) 
		return;

	FILE *fp = qfopen(filename, "rb");
	if (fp == NULL) 
	{
		error("COFF Plugin -> Error: Can't open %s file for reading\n", filename);
		return;
	}

	segment_t * seg = getnseg(0);
	if (seg != NULL) 
		eaStart = seg->startEA;
	
	seg = getnseg(get_segm_qty()-1);

	if (seg != NULL) 
		eaEnd = seg->endEA;

	msg("COFF Plugin -> file: '%s'\n", filename);
	msg("COFF Plugin -> start addr: %08a, end addr: %08a\n", eaStart, eaEnd);

	try
	{
		show_wait_box("'%s'\n COFF Plugin: symbols loading...", filename);
		clearBreak();

		if (stristr(filename, ".obj") != NULL)	
			applyCoffObject(fp);
		else
		{
			if ((stristr(filename, ".lib") != NULL) && LibSignatureOk(fp))
				ParseLibFile(fp);
		}
	}
	catch (std::exception & e)
	{
		msg("Exception %s in the main function\n",e.what());
	}
	catch (...)
	{
		msg("Unknown Exception in the main function\n");
	}

	hide_wait_box();    
	qfclose(fp);

	refresh_idaview_anyway();
}

//--------------------------------------------------------------------------
char comment[] = "COFF obj/lib signature loader...";

char help[] = "COFF signature loader plugin\n\nwritten by gokhan erbay, Hex and others\n";


//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = "COFF object signature loader";


// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = "Ctrl+1";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
	IDP_INTERFACE_VERSION,
	0,                    // plugin flags
	init,                 // initialize

	term,                 // terminate. this pointer may be NULL.

	run,                  // invoke plugin

	comment,              // long comment about the plugin
	// it could appear in the status line
	// or as a hint

	help,                 // multiline help about the plugin

	wanted_name,          // the preferred short name of the plugin
	wanted_hotkey         // the preferred hotkey to run the plugin
};
